/*
  BLOCK.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/


#include "misc.h"
#include "casm.h"
#include "block.h"
#include "parser.h"
#include "numbers.h"
#include "tables.h"
#include "alias.h"
#include "tokenize.h"
#include "numbers.h"

/*************************** GENERAL DEFINITIONS ***************************/

TBlock::TBlock(const char *name) {
  Name = name;
  BlockType = UNDEFINED;
  DefinedYet = FALSE;
  ObjCode = 0;
  ObjCodeSize = 0;
  OffsetInDestFile = 0;
}

THash BlockList;

THash NearJumpList;


unsigned NextTempLabel;  // number of the next temp label to be read

//--------------------------------------------------------------------------
FString GenerateTempLabel(unsigned num) {
//--------------------------------------------------------------------------
  char buf[8];
  char *c = buf;

  *c = '~'; c++;

  *c = 'a' + (num & 15);
  num /= 16; c++;
  *c = 'a' + (num & 15);
  num /= 16; c++;
  *c = 'a' + (num & 15);
  num /= 16; c++;
  *c = 'a' + (num & 15);

  c++;  *c = 0;

  return buf;
}

//--------------------------------------------------------------------------
void ReadLabelDef(TASMFileTokenizer& ftok,TBlock *Block,TToken& Name) {
//--------------------------------------------------------------------------
/*
  This reads a line label definition.  Unlike ReadAssemblyInstr(),
  ReadLineLabel() does not return a BOOL indicating success; it expects a
  line label.  This was necessary because line labels can only be
  identified by the succeeding ':' token, which would could only be
  identified by peeking ahead two tokens, and the tokenizer doesn't
  currently support this.
*/
  TToken cur;

  if (Name == '&') cur.Text = GenerateTempLabel(NextTempLabel++);
  else {
    cur = ftok.GetNextToken();

    if (cur != ':') {
      // not a line label; who knows what it is
      cur.WriteFileLine(ErrStr);

      // be nice to the user
      if (IsReservedWord(Name.Text)) {
	Name.Text.AllUpper();
	ErrStr << '"' << Name.Text << "\" may not be used here" << ENDE;
      }

      if (IsReservedWord(cur.Text)) {
	cur.Text.AllUpper();
	ErrStr << '"' << cur.Text << "\" may not be used here" << ENDE;
      }

      ErrStr << "Unknown symbol: " << cur.Text << ENDE;
    }
    cur = Name;

    // make sure the label name is available:
    if (IsNameInUse(cur.Text)) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "The identifier \"" << cur.Text << "\" is already in use" << ENDE;
    }

    // make sure there's not already a label with this name:
    if (Block->LabelDefs.Find(cur.Text.c_str())) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "The label \"" << cur.Text << "\" has already been defined" << ENDE;
    }
  }

  // add this to the definitions:
  TBlock::TLabelDef *NewDef = new TBlock::TLabelDef(cur.Text,
    Block->ObjCodeSize);
  if (!NewDef) ErrStr.OutOfMemory();

  Block->LabelDefs.Insert(NewDef);
}

//--------------------------------------------------------------------------
TBlock::TLabelRef *ReadLabelReference(TASMFileTokenizer& ftok,
  TBlock *BlockContainingReference,TToken& Name) {
//--------------------------------------------------------------------------
/*
  Reads a label reference, returning a TLabelRef structure.  The calling
  function must fill in the OffsetInBlock and RefType fields.
*/

  TBlock::TLabelRef *ret = new TBlock::TLabelRef;
  if (!ret) ErrStr.OutOfMemory();

  TToken cur = Name;
  FString NameOfBlockWhereDefined;
  TToken LabelName;

  unsigned TempNum = NextTempLabel;

  if (Name == '+') {
    while (ftok.PeekNextToken() == '+') {
      ftok.GetNextToken();
      TempNum++;
    }
    cur.Text = GenerateTempLabel(TempNum);
  } else if (Name == '-') {
    TempNum--;
    while (ftok.PeekNextToken() == '-') {
      ftok.GetNextToken();
      TempNum--;
    }
    cur.Text = GenerateTempLabel(TempNum);
  }

  ret->Filename = cur.Filename;
  ret->LineNum = cur.LineNum;

  if (ftok.PeekNextToken() != '|') {
    // assume it's defined in this block:
    LabelName = cur;
    NameOfBlockWhereDefined = BlockContainingReference->Name;
  } else {
    // read the '|'
    NameOfBlockWhereDefined = cur.Text;
    ftok.GetNextToken();
    LabelName = ftok.GetNextToken();
  }

  ret->DefBlock = (TBlock *)BlockList.Find(NameOfBlockWhereDefined.c_str());
  if (!ret->DefBlock) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "Unknown block \"" << NameOfBlockWhereDefined << '"' << ENDE;
  }

  if ((LabelName.Text[0] != '~') && !IsID(LabelName.Text)) {
    LabelName.WriteFileLine(ErrStr);
    ErrStr << "Invalid label name: \"" << LabelName << '"' << ENDE;
  }
  ret->Name = LabelName.Text;

  ret->AddAmount = 0;

  return ret;
}

//--------------------------------------------------------------------------
void ReadAnyLocalAliases(TASMFileTokenizer& ftok,TBlock *block) {
//--------------------------------------------------------------------------
  TToken tok;

  BOOL IsWord;
  BOOL IsConst = FALSE;
  TAlias *newa;

  while (1) {
    tok = ftok.PeekNextToken();

    if (tok == "ALIASB") IsWord = FALSE;
    else if (tok == "ALIASW") IsWord = TRUE;
    else if (tok == "CONSTB") { IsWord = FALSE; IsConst = TRUE; }
    else if (tok == "CONSTW") { IsWord = TRUE;  IsConst = TRUE; }
    else if (tok == "VB") IsWord = FALSE;
    else if (tok == "VW") IsWord = TRUE;
    else return;

    ftok.GetNextToken();

    if ((tok.Text[0] == 'v') || (tok.Text[0] == 'V'))
      newa = ReadAutoAliasDef(ftok,IsWord);
    else newa = ReadAliasDef(ftok,IsConst,IsWord,block);

    block->LocalAliases.Insert(newa);
  }
}


/************************** CODE-SPECIFIC DEFINITIONS **********************/

//--------------------------------------------------------------------------
int FindByteForInstr(instrtype instr,paramtype& param) {
//--------------------------------------------------------------------------
  int code = 0;
  do {
    if (instrtable[code].instr == instr) {
      if (instrtable[code].param == param) return code;

      // if match near jumps to far jumps, changing param:
      if ((param == TYPEK) || (param == TYPEG))
	if ((instrtable[code].param == TYPEK)
	 || (instrtable[code].param == TYPEG)) {
	   param = instrtable[code].param;
	   return code;
	}
    }
  } while (++code < 256);
  return -1;
}

//--------------------------------------------------------------------------
BOOL ReadAssemblyInstr(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
#define INSTR_DAT 	-1
#define INSTR_DATW	-2

  TToken cur = ftok.PeekNextToken();

  TByteSize ParamSize = BYTE;  // byte, lobyte, hibyte, word
  unsigned Param;       // actual parameter data (if any)
  paramtype ParamType;  // TYPEA...TYPEK
  TAlias *CurAlias; // current alias being read
  TBlock::TLabelRef *LabelRef = 0; // if a label is referenced, this points
				   // to the reference record

  // translate extended instructions
  if (cur.Text == "???") cur.Text = "NOP"; // treat as NOP

  // Translate "JEQ $1000" to "BNE + JMP $1000 &", etc.
  if (cur.Text == "JEQ")
    Block->ObjCode[Block->ObjCodeSize++] = 0xd0;
  else if (cur.Text == "JNE")
    Block->ObjCode[Block->ObjCodeSize++] = 0xf0;
  else if (cur.Text == "JCC")
    Block->ObjCode[Block->ObjCodeSize++] = 0xb0;
  else if (cur.Text == "JCS")
    Block->ObjCode[Block->ObjCodeSize++] = 0x90;
  else if (cur.Text == "JPL")
    Block->ObjCode[Block->ObjCodeSize++] = 0x30;
  else if (cur.Text == "JMI")
    Block->ObjCode[Block->ObjCodeSize++] = 0x10;
  else if (cur.Text == "JVC")
    Block->ObjCode[Block->ObjCodeSize++] = 0x70;
  else if (cur.Text == "JVS")
    Block->ObjCode[Block->ObjCodeSize++] = 0x50;
  else goto NotAJBranch;
    Block->ObjCode[Block->ObjCodeSize++] = 0x03;
    cur.Text = "JMP";
  NotAJBranch:;


  // handle regular instructions
  int InstrCode;
  int InstrByte = -1;

  if ((cur.Text == "DAT") || (cur.Text == "BYT")) { 	// treat specially
    InstrCode = INSTR_DAT;
    Block->ObjCodeSize--;  // force labels to match up
  } else if (cur.Text == "DATW") {
    InstrCode = INSTR_DATW;
    Block->ObjCodeSize--;
  } else {
    InstrCode = GetInstrNum(cur.Text.c_str());
    if (InstrCode < 0) return FALSE; // not an instruction
  }

  ftok.GetNextToken(); // read the instruction

  // find out if it has parameters:
  ParamType = TYPEA;  // assume no parameters
  if ((InstrCode != INSTR_DAT) && (InstrCode != INSTR_DATW))
    InstrByte = FindByteForInstr((instrtype) InstrCode,ParamType);

  switch ((instrtype) InstrCode) {
    case ASL:
    case LSR:
    case ROR:
    case ROL:
      if (ftok.PeekNextToken() == "ACC") {
	ftok.GetNextToken();

	// it's a one-byte, no arguments instr.
	Block->ObjCode[Block->ObjCodeSize++] = InstrByte;
	return TRUE;
      }
      break;
    default:
      if (InstrByte >= 0) {
	// it's a one-byte, no arguments instr.
	Block->ObjCode[Block->ObjCodeSize++] = InstrByte;
	return TRUE;
      }
  }

  // it must have arguments; read them
  BOOL StartsWithLP = FALSE;  // starts with (
  BOOL StartsWithNum = FALSE;  // starts with #
  Param = 0;  // start at zero

  cur = ftok.PeekNextToken();
  if (cur == '#') {
    StartsWithNum = TRUE;
    ftok.GetNextToken();
    cur = ftok.PeekNextToken();
  } else if (cur == '(') {
    StartsWithLP = TRUE;
    ftok.GetNextToken();
    cur = ftok.PeekNextToken();
  }

  // check for > and < prefixes
  if (cur.Text == '>') {
    ParamSize = HIBYTEOFWORD;
    ftok.GetNextToken();
    cur = ftok.PeekNextToken();
  } else if (cur.Text == '<') {
    ParamSize = LOBYTEOFWORD;
    ftok.GetNextToken();
    cur = ftok.PeekNextToken();
  }

/*

  [<|>] [Label|Alias|Num] [<+|-> <Alias|Num>] [<+|-> <Alias|Num>] ...

  e.g. 3, >label, <alias + 3 - alias - %11000101, >$1234, <label + 'A', etc.

*/
  // loop through, reading numbers and aliases being added/subtracted:
  BOOL SubtractIt = FALSE; // whether the stuff being read is after a '-'
  unsigned newval;
  BOOL FirstTimeThroughLoop = TRUE;

  while(1) {
    cur = ftok.PeekNextToken();
    switch (ReadByteOrWord(ftok,newval)) {
      case 0: // nothing read
	// perhaps it's an alias or label
	if (isidchar(cur.Text[0]) || (cur == '+') || (cur == '-')) {
	  // check for an alias:

	  CurAlias = (TAlias *)Block->LocalAliases.Find(cur.Text.c_str());
	  if (!CurAlias)
	    CurAlias = (TAlias *)AliasList.Find(cur.Text.c_str());

	  if (CurAlias) {
	    ftok.GetNextToken(); // read the peeked text

	    if (CurAlias->Attrs.IsConst && !StartsWithNum) {
	      cur.WriteFileLine(ErrStr);
	      ErrStr << "Constant used without '#'" << ENDW;
	    }

	    // this is for "unused alias" warnings
	    CurAlias->Attrs.IsUsed = TRUE;

	    if (SubtractIt) Param -= CurAlias->Data;
	    else Param += CurAlias->Data;

	    if ((ParamSize == BYTE) && CurAlias->Attrs.IsWord)
	      ParamSize = WORD;
	  } else {
	    // it's not an alias; it can only be a line label:

	    // are line labels allowed here?
	    if (!FirstTimeThroughLoop) {
	      cur.WriteFileLine(ErrStr);
	      ErrStr << "Unknown alias or misplaced line label: \""
		<< cur << '"' << ENDE;
	    }

	    LabelRef = ReadLabelReference(ftok,Block,ftok.GetNextToken());
	    LabelRef->OffsetInBlock = Block->ObjCodeSize + 1; // skip instr
	    if (ParamSize == BYTE) ParamSize = WORD;
	  }
	} else {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "Syntax error: \"" << cur << '"' << ENDE;
	}
	break;
      case 1: // word
	if (ParamSize == BYTE) ParamSize = WORD;
      case -1: // byte (and word)
	if (SubtractIt) Param -= newval;
	else Param += newval;
    }

    FirstTimeThroughLoop = FALSE;

    cur = ftok.PeekNextToken();
    if (cur == '-') {
      ftok.GetNextToken();  // read the '-'
      SubtractIt = TRUE;
      continue;
    } else if (cur == '+') {
      ftok.GetNextToken();  // read the '+'
      SubtractIt = FALSE;
      continue;
    }

    // otherwise, quit the loop
    break;
  }

  // next, figure out which type it is:

  // TYPED = #$00
  if (StartsWithNum) {
    if (ParamSize == WORD) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Cannot load a 16-bit value" << ENDE;
    }
    ParamType = TYPED;
    goto StoreIt;
  }

  if (ParamSize != WORD) {
    if (StartsWithLP) {  // TYPEB = ($00,X) / TYPEE = ($00),Y
      if (ftok.PeekNextToken() == ',') {
	ftok.GetNextToken(); // read the ','
	if ((ftok.GetNextToken() != 'X') || (ftok.GetNextToken() != ')')) {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "\",X)\" expected" << ENDE;
	}
	ParamType = TYPEB;
	goto StoreIt;
      } else {
	if ((ftok.GetNextToken() != ')')
	 || (ftok.GetNextToken() != ',')
	 || (ftok.GetNextToken() != 'Y')) {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "\"),Y\" expected" << ENDE;
	}
	ParamType = TYPEE;
	goto StoreIt;
      }
    }

    // doesn't start with (

    if (ftok.PeekNextToken() == ',') { // TYPEF = $00,X / TYPEL = $00,Y
      ftok.GetNextToken(); // read the ','
      cur = ftok.GetNextToken();
      if (cur == 'X') {
	ParamType = TYPEF;
	goto StoreIt;
      } else if (cur == 'Y') {
	ParamType = TYPEL;
	goto StoreIt;
      } else {
	cur.WriteFileLine(ErrStr);
	ErrStr << "\",X\" or \",Y\" expected" << ENDE;
      }
    }

    // TYPEC = $00
    ParamType = TYPEC;
    goto StoreIt;
  } else { // it's a word
    if (StartsWithLP) { // TYPEJ = ($0000)
      if (ftok.GetNextToken() != ')') {
	cur.WriteFileLine(ErrStr);
	ErrStr << "')' expected" << ENDE;
      }
      ParamType = TYPEJ;
      goto StoreIt;
    }

    if (ftok.PeekNextToken() == ',') { // TYPEH = $0000,Y / TYPEI = $0000,X
      ftok.GetNextToken();
      cur = ftok.GetNextToken();
      if (cur == 'X') {
	ParamType = TYPEI;
	goto StoreIt;
      } else if (cur == 'Y') {
	ParamType = TYPEH;
	goto StoreIt;
      } else {
	cur.WriteFileLine(ErrStr);
	ErrStr << "\",X\" or \",Y\" expected" << ENDE;
      }
    }

    // TYPEG = $0000
    ParamType = TYPEG;
    goto StoreIt;
  }

  StoreIt:;
  // This also changes branches to TYPEK
  if (InstrCode == INSTR_DAT) {
    if (ParamSize == WORD) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "DAT requires an 8-bit value" << ENDE;
    }
    Block->ObjCodeSize++;  // (changed at beginning)
  } else if (InstrCode == INSTR_DATW) {
    Block->ObjCodeSize++;  // (changed at beginning)
  } else {
    InstrByte = FindByteForInstr((instrtype) InstrCode,ParamType);

    if (InstrByte < 0) {
      // wrong type of parameters:

      cur.WriteFileLine(ErrStr);
      ErrStr << instrstr[InstrCode] << " can't take ";
      switch (ParamType) {
	case TYPEB: ErrStr << "($00,X)"; break;
	case TYPEC: ErrStr << "$00"; break;
	case TYPED: ErrStr << "#$00"; break;
	case TYPEE: ErrStr << "($00),Y"; break;
	case TYPEF: ErrStr << "$00,X"; break;
	case TYPEL: ErrStr << "$00,Y"; break;
	case TYPEK: ErrStr << "$0000"; break;
	case TYPEG: ErrStr << "$0000"; break;
	case TYPEH: ErrStr << "$0000,Y"; break;
	case TYPEI: ErrStr << "$0000,X"; break;
	case TYPEJ: ErrStr << "($0000)"; break;
	default: ErrStr << "*** unknown param type ***" << PROGBUG;
      }
      ErrStr << "-type parameters" << ENDE;
    }

    // now store the instruction:
    Block->ObjCode[Block->ObjCodeSize++] = (char) InstrByte;
  }

  // if there is a label reference, store it
  if (LabelRef) {

    // set LabelRef->RefType
    if (ParamType == TYPEK) { // BEQ
      if (ParamSize != WORD) {
	// write the file / line num
	TToken tok;
	tok.Filename = LabelRef->Filename;
	tok.LineNum = LabelRef->LineNum;
	tok.WriteFileLine(ErrStr);

	ErrStr << "16-bit parameter required for near jumps" << ENDE;
      }
      LabelRef->RefType = TBlock::TLabelRef::NEAR;
      Block->ObjCode[Block->ObjCodeSize++] = 0;
    } else {
      switch (ParamSize) {
	case WORD:
	  LabelRef->RefType = TBlock::TLabelRef::FAR;
	  Block->ObjCode[Block->ObjCodeSize++] = 0;
	  Block->ObjCode[Block->ObjCodeSize++] = 0;
	  break;
	case HIBYTEOFWORD:
	  LabelRef->RefType = TBlock::TLabelRef::HI;
	  Block->ObjCode[Block->ObjCodeSize++] = 0;
	  break;
	case LOBYTEOFWORD:
	  LabelRef->RefType = TBlock::TLabelRef::LO;
	  Block->ObjCode[Block->ObjCodeSize++] = 0;
	  break;
	case BYTE:
	default:
	  ftok.WriteFileLine(ErrStr);
	  ErrStr << "Invalid ParamSize" << PROGBUG;
      }
    }

    // Param becomes the AddAmount
    LabelRef->AddAmount = Param;

    // store it:
    Block->LabelRefs.Insert(LabelRef);
  } else {
    // if there isn't a label reference...
    if (ParamType == TYPEK) {
      if (ParamSize != WORD) {
	cur.WriteFileLine(ErrStr);
	ErrStr << "Address expected" << ENDE;
      }

      TNearJump *nj = new TNearJump(Block,Block->ObjCodeSize,
	Param,cur.Filename,cur.LineNum);
      if (!nj) ErrStr.OutOfMemory();
      NearJumpList.Insert(nj);
      Block->ObjCode[Block->ObjCodeSize++] = 0;
    } else {
      switch (ParamSize) {
	case HIBYTEOFWORD:
	  Block->ObjCode[Block->ObjCodeSize++] = (char) (Param >> 8);
	  break;
	case BYTE:
	case LOBYTEOFWORD:
	  Block->ObjCode[Block->ObjCodeSize++] = (char) Param;
	  break;
	case WORD:
	  Block->ObjCode[Block->ObjCodeSize++] = (char) Param;
	  Block->ObjCode[Block->ObjCodeSize++] = (char) (Param >> 8);
	  break;
	default:
	  ErrStr << "Invalid ParamType" << PROGBUG;
      }
    }
  }

  return TRUE;
}

/************************** DATA-SPECIFIC DEFINITIONS **********************/

//--------------------------------------------------------------------------
void ReadBitmap(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  int ch;
  unsigned curbyte;
  int bitcount;

  ftok.SkipWhite();

  // read eight '.' or 'X':
  curbyte = 0;
  bitcount = 0;
  do {
    ftok.SkipWhite();

    curbyte <<= 1;
    ch = ftok.Get();

    switch (ch) {
      case '.':
	++bitcount;
	break;
      case 'X':
      case 'x':
      case '*':
	curbyte |= 1;
	++bitcount;
	break;
      case ']':
	if (bitcount != 0) {
	  ftok.WriteFileLine(ErrStr);
	  ErrStr << "Number of bits is not a multiple of 8" << ENDE;
	}
	return;
      case EOF:
	ErrStr << "Unexpected end of file in middle of DATA block"
	  << ENDE;
	break;
      default:
	ftok.WriteFileLine(ErrStr);
	ErrStr << "Unknown character: ";
	ErrStr.QuoteCharacter(ch);
	ErrStr << ENDE;
    }

    if (bitcount == 8) {
      Block->ObjCode[Block->ObjCodeSize++] = curbyte;
      bitcount = 0;
    }
  } while(1);
}

//--------------------------------------------------------------------------
void ReadDataNumber(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  unsigned num;

  switch(ReadByteOrWord(ftok,num)) {
    case -1: // byte
      Block->ObjCode[Block->ObjCodeSize++] = num;
      break;
    case 1: // word
      Block->ObjCode[Block->ObjCodeSize++] = num;
      Block->ObjCode[Block->ObjCodeSize++] = num >> 8;
      break;
    case 0:
      ftok.WriteFileLine(ErrStr);
      ErrStr << "8-bit or 16-bit number expected" << ENDE;
      break;
  }
}

// Pet ASCII
//--------------------------------------------------------------------------
void ReadPAString(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  ftok.GetNextToken(); // read the '
  DString str = ftok.ReadToDelim("\"\n");

  if (ftok.GetNextToken() != '"') {
    ftok.WriteFileLine(ErrStr);
    ErrStr << "'\"' expected" << ENDE;
  }

  const char *src = str.c_str();
  char *dest = Block->ObjCode + Block->ObjCodeSize;

  while (*src) {
    *dest = ASCII2PETASCII(*src);
    dest++;
    src++;
    Block->ObjCodeSize++;
  }
};

// Screen Code
//--------------------------------------------------------------------------
void ReadSCString(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  ftok.GetNextToken(); // read the '
  DString str = ftok.ReadToDelim("\'\n");

  if (ftok.GetNextToken() != '\'') {
    ftok.WriteFileLine(ErrStr);
    ErrStr << "\"'\" expected" << ENDE;
  }

  const char *src = str.c_str();
  char *dest = Block->ObjCode + Block->ObjCodeSize;

  while (*src) {
    *dest = ASCII2ScreenCode(*src);
    dest++;
    src++;
    Block->ObjCodeSize++;
  }
};

//--------------------------------------------------------------------------
void ReadDataID(TBlock *Block,TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  TToken cur = ftok.GetNextToken();

  // first, assuming it's a label, check for < or >
  TBlock::TLabelRef::TRefType RefType = TBlock::TLabelRef::FAR;	// assume FAR

  if (cur == '<') {
    RefType = TBlock::TLabelRef::LO;
    cur = ftok.GetNextToken();
  } else if (cur == '>') {
    RefType = TBlock::TLabelRef::HI;
    cur = ftok.GetNextToken();
  } else if (cur == '^') {
    RefType = TBlock::TLabelRef::NEAR;
    cur = ftok.GetNextToken();
  }

  // if it's a near label reference, we can skip the other checks
  if (RefType != TBlock::TLabelRef::NEAR) {

    // Check for an alias:
    TAlias *TestAlias = (TAlias *)Block->LocalAliases.Find(cur.Text.c_str());
    if (!TestAlias) TestAlias = (TAlias *)AliasList.Find(cur.Text.c_str());
    if (TestAlias) {
      TestAlias->Attrs.IsUsed = TRUE;  // this is for "unused alias" warnings

      if (ftok.PeekNextToken() == ':') {
	cur.WriteFileLine(ErrStr);
	ErrStr << "Label name is already being used by an ALIAS" << ENDE;
      }

      // if '<' or '>' was used, then write the appropriate part of the alias
      if (RefType != TBlock::TLabelRef::FAR) {
	if (!TestAlias->Attrs.IsWord) {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "16-bit value needed for '<' and '>'" << ENDE;
	}

	if (RefType == TBlock::TLabelRef::LO)
	  Block->ObjCode[Block->ObjCodeSize++] = (char) TestAlias->Data;
	else Block->ObjCode[Block->ObjCodeSize++] = (char) (TestAlias->Data >> 8);
      } else {
	// otherwise just write the normal alias:

	Block->ObjCode[Block->ObjCodeSize++] = (char) TestAlias->Data;
	// if it's a word, write the MSB:
	if (TestAlias->Attrs.IsWord) {
	  Block->ObjCode[Block->ObjCodeSize++] = (char) (TestAlias->Data >> 8);
	}
      }

      return;
    }

    // is it a label definition?
    if ((ftok.PeekNextToken() == ':') || (cur == '&')) {
      if (RefType != TBlock::TLabelRef::FAR) {
	cur.WriteFileLine(ErrStr);
	ErrStr << "Syntax error: '<' or '>' cannot precede a label definition" << ENDE;
      }

      ReadLabelDef(ftok,Block,cur);
      return;
    }

  }

  // it must be a reference:
  TBlock::TLabelRef *ref = ReadLabelReference(ftok,Block,cur);

  ref->OffsetInBlock = Block->ObjCodeSize;
  ref->RefType = RefType;

  Block->ObjCodeSize++;
  if (RefType == TBlock::TLabelRef::FAR) Block->ObjCodeSize++;
  Block->LabelRefs.Insert(ref);
}

/************************* BLOCK-READING DEFINITIONS ***********************/

//--------------------------------------------------------------------------
TBlock *ReadBlockHeader(TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
/*
  Reads the name of the block and finds the corresponding BlockList entry.
*/
  TToken BlockName = ftok.GetNextToken();
  if (!IsID(BlockName.Text)) {
    BlockName.WriteFileLine(ErrStr);
    ErrStr << "Invalid block name" << ENDE;
  }

  return (TBlock *)BlockList.Find(BlockName.Text.c_str());
}

//--------------------------------------------------------------------------
void ReadBlock(TASMFileTokenizer& ftok,TBlock::TBlockType InitialMode) {
//--------------------------------------------------------------------------
  TBlock *Block = ReadBlockHeader(ftok);

  if (!Block) {
    // this block wasn't listed in the layout section
    if (InitialMode == TBlock::CODE) {
      if (!ftok.SkipTo("ENDC"))
	ErrStr << "File ended in middle of CODE block" << ENDE;
      return;
    } else {
      if (!ftok.SkipTo("ENDD"))
	ErrStr << "File ended in middle of DATA block" << ENDE;
      return;
    }
  }

  Block->BlockType = InitialMode;
  Block->ObjCode = (char far *) (GP.ObjBuf + GP.ObjBufSize);
  TToken cur;
  TBlock::TBlockType CurMode = InitialMode;
  NextTempLabel = 0;

  ReadAnyLocalAliases(ftok,Block);

  do {
    cur = ftok.PeekNextToken();
    if (cur == "" || ftok.Eof())
      ErrStr << "Unexpected end of file in middle of "
	<< ((InitialMode == TBlock::CODE) ? "CODE" : "DATA")
	<< " block" << ENDE;

    if (cur == ((InitialMode == TBlock::CODE) ? "ENDC" : "ENDD")) {
      ftok.GetNextToken(); // read the ENDS/ENDD
      Block->DefinedYet = TRUE;
      GP.ObjBufSize += Block->ObjCodeSize + 3;  // safety space
      return;
    }

    if (cur == "DATA") {
      ftok.GetNextToken();  //  read the "DATA"
      CurMode = TBlock::DATA;
      continue;
    }

    if (cur == "CODE") {
      ftok.GetNextToken();  //  read the "CODE"
      CurMode = TBlock::CODE;
      continue;
    }

    if (CurMode == TBlock::CODE) {
      if (ReadAssemblyInstr(Block,ftok)) continue;

      ReadLabelDef(ftok,Block,ftok.GetNextToken());
    } else {
      if (isdigit(cur.Text[0]) || (cur == '$')
	|| (cur == '%')) ReadDataNumber(Block,ftok);
      else if (cur == '\"') ReadPAString(Block,ftok);
      else if (cur == '\'') ReadSCString(Block,ftok);
      else if (cur == '[') {
	ftok.GetNextToken();
	ReadBitmap(Block,ftok);
      } else if (IsID(cur.Text) || (cur == '>')
	|| (cur == '<') || (cur == '&')
	|| (cur == '-') || (cur == '+')
	|| (cur == '^')) ReadDataID(Block,ftok);
      else {
	cur.WriteFileLine(ErrStr);

	if (IsReservedWord(cur.Text)) {
	  cur.Text.AllUpper();
	  ErrStr << '"' << cur.Text << "\" may not be used here" << ENDE;
	}

	ErrStr << "Syntax error: " << cur.Text << ENDE;
      }
    }
  } while (1);
}
