/* 
  TOKENIZE.H - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#ifndef _TOKENIZE_H
#define _TOKENIZE_H

#include <fstream.h>

#include "general.h"
#include "petestr.h"

struct TToken {
  FString Text;
  FString Filename;
  unsigned long LineNum;

  TToken(const FString& _Text,const FString& _Filename = "",
    unsigned long _LineNum = 0)
    : Text(_Text), Filename(_Filename) { LineNum = _LineNum; }
  TToken() { LineNum = 0; }

  TToken& operator =(TToken &newval) {
    Text = newval.Text;
    Filename = newval.Filename;
    LineNum = newval.LineNum;
    return *this;
  }

  BOOL operator ==(const char *str) { return Text == str; }
  BOOL operator ==(const char c) { return Text == c; }
  BOOL operator ==(const FString& s) { return Text == s; }
  BOOL operator ==(const TToken& t) { return Text == t.Text; }
  BOOL operator !=(const char *str) { return Text != str; }
  BOOL operator !=(const char c) { return Text != c; }
  BOOL operator !=(const FString& s) { return Text != s; }
  BOOL operator !=(const TToken& t) { return Text != t.Text; }

  BOOL isnull() { return !Text.Length(); }

  TErrStr& WriteFileLine(TErrStr& es) {
    return es << '[' << Filename << ':' << LineNum << "] ";
  }
};

#include "diskacc.h"

#define BUFSIZE 4096

class pgistream {
  char buf[BUFSIZE+4];
  int curp;
  int buftop;
  int peekch;
  BOOL peeked;
  BOOL EOFhit;

  enum {
    FILEOPEN,
    FILECLOSED,
    ERROR,
    FILEEOF
  } status;

  FString fname;

  unsigned handle;

public:
  pgistream() {
    peeked = FALSE;
    curp = 0;
    buftop = 0;
    peekch = 0;
    EOFhit = FALSE;
    status = FILECLOSED;
  }

  ~pgistream() {
    if (status == FILEOPEN) {
      fasdos_close(handle);
    }
  }

  void open(const char *_fname) {
    fname = _fname;
    curp = 0;

    if ((status == FILEOPEN) || (status == FILEEOF)) fasdos_close(handle);

    handle = fasdos_open(fname.c_str());

    if (handle == 0xffff) status = ERROR;
    else status = FILEOPEN;
  }

  int get() {
    if (status != FILEOPEN) return EOF;
    if (peeked) {
      peeked = FALSE;
      return peekch;
    }
    if (curp < buftop) return buf[curp++];

    // overflow:
    unsigned t = fasdos_read(handle,buf,BUFSIZE);
    if (t == 0xffff) {
      status = ERROR;
      return EOF;
    }

    curp = 0;
    buftop = t;

    return buf[curp++];
  }

  int peek() {
    if (peeked)  return peekch;
    peeked = TRUE;
    peekch = get();
    return peekch;
  }

  void putback(int ch) {
    if (peeked) ErrStr << "stream problems" << PROGBUG;
    peekch = ch;
    peeked = TRUE;
  }

  BOOL fail() { return status == ERROR; }

  void clear() { }


};


ostream& operator <<(ostream& os,TToken t);

class TASMFileTokenizer {
protected:
  struct TFileData {
    pgistream is;
    DString Filename;
    long LineNum;

    int GetAndCountLines() {
      int ch = is.get();
      if (ch == '\n')
	LineNum++;
      return ch;
    }
    void PutBackAndCountLines(int ch) {
      if (ch == '\n') LineNum--;
      is.putback(ch);
    }
    BOOL operator ==(TFileData &d) {
      return &d == this;
    }

    TErrStr& WriteFileLine(TErrStr& es) {
      return es << '[' << Filename << ':' << LineNum << "] ";
    }

    TFileData *Next;
  };

  TFileData *FileStack;

  // these do not support peeking:
  BOOL _Eof();
  int  _GetWithComment();  // pulls a character from file, processing
		// "@" directives, and counting lines
  int  _Get();  // gets, removing comments

  TToken PeekToken;    // used by PeekNextToken()
  BOOL TokenPeeked;

  int PeekChar;
  BOOL CharPeeked;

public:
  TASMFileTokenizer(const DString& startfile) {
    FileStack = 0;
    TokenPeeked = FALSE;
    CharPeeked = FALSE;
    Open(startfile);
  }
  ~TASMFileTokenizer();

  void Open(const DString& Filename) {  // push a new file on the stack
    TFileData *d = new TFileData;
    d->LineNum = 1;
    d->Filename = Filename;
    d->Filename.AllUpper();
    d->is.open(d->Filename.c_str());
    if (d->is.fail())
      ErrStr << "Unable to open \"" << d->Filename.c_str() << '\"' << ENDE;
    d->is.clear();    // prevent false alarms

    // push this on the stack:
    d->Next = FileStack;
    FileStack = d;
  }

  BOOL Eof() { if (CharPeeked | TokenPeeked) return FALSE;
	       else return _Eof(); }
  int GetWithComment();
  int Get();
  int PeekWithComment();
  int Peek();

  BOOL IsWhite(int c) { return isspace(c); }

  BOOL IsSymbol(int c);

  void SkipWhite();

  TToken GetNextToken();
  TToken PeekNextToken();

  BOOL SkipTo(const FString& token);  // returns FALSE if EOF encountered

  // Reads up to until it reaches one of the characters in delim.  Does not extract that character
  DString ReadToDelim(const char *delim);

  TErrStr& WriteFileLine(TErrStr& es) {
    return es << '[' << FileStack->Filename << ':' << FileStack->LineNum << "] ";
  }

protected:
  TToken _GetNextToken();  // doesn't support peeking
};

#endif
