/* 
  PETESTR.H - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    This defines two types of string objects, DString and FString.
  DStrings are allocated dynamically, and resize themselves to the
  length of the data.  FStrings have a fixed length MAXFIXSTRING,
  but are much faster.  Both types support basic operations for
  concatenation, scanning, etc.

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#ifndef _PETESTR_H
#define _PETESTR_H

#include "general.h"

#include <iostream.h>
#include <alloc.h>
#include <string.h>
#include <ctype.h>


// super fast string compare:
int fastcmp(const char far *s1,const char far *s2);

class FString; // forward declaration

/******************************************************************
			     Variable-Length
******************************************************************/

class DString {
friend ostream& operator << (ostream& os,DString& s);
protected:
  char far *Str;
  size_t _Length;
  size_t BufSize;
  int delta;

  // resizes buffer to hold a string of size _Length:
  void Resize();

public:
  DString(const char far *src = "");
  DString(const char c);
  DString(const DString& src);
  DString(const FString& src);
  ~DString() {
    delete Str;
  }
  const char far *c_str() const { return Str; }
  int Length() const { return _Length; }

  DString& operator =(const DString& src);
  DString& operator =(const char far *src);
  DString& operator =(const char c);
  BOOL operator ==(const char c) const {
    if (_Length != 1) return FALSE;
    return toupper(Str[0]) == toupper(c);
  }
  BOOL operator == (const DString& src) const {
    if (Length() != src.Length()) return FALSE;
    // return !stricmp(src.Str,Str);
    return fastcmp(src.Str,Str);
  }
  BOOL operator == (const char far *src) const {
    // return !stricmp(src,Str);
    return fastcmp(src,Str);
  }
  BOOL operator !=(const char c) const { return !(*this == c); }
  BOOL operator !=(const DString& src) const { return !(*this == src); }
  BOOL operator !=(const char far *src) const { return !(*this == src); }

  DString& operator +=(const DString& src) {
    _Length += src._Length;
    Resize();
    strcat(Str,src.Str);
    return *this;
  }

  DString operator +=(const char far *src) {
    _Length += strlen(src);
    Resize();
    strcat(Str,src);
    return *this;
  }

  DString operator +=(const char c) {
    _Length++;
    Resize();
    Str[_Length-1] = c;
    Str[_Length] = 0;
    return *this;
  }

  DString operator +(const DString& src);
  DString operator +(const char far *src);
  DString operator +(const char c);

  int NumLines();
  void ShrinkTo(int MaxLen);
  void Insert(char c,int pos=0);
  void Delete(int pos);
  char operator[] (int pos) {
    if ((pos < 0) || (pos > _Length))
      ErrStr << "Invalid string index" << ENDE;
    return Str[pos];
  }
  void SetDelta(int _delta) {
    if (_delta > 0) delta = _delta;
  }

  void AllUpper() { strupr(Str); }
  void AllLower() { strlwr(Str); }

  // removes everything to the right of the last period
  void RemoveFileExtension();
};

ostream& operator << (ostream& os,DString& s);

/******************************************************************
			     Fixed-Length
******************************************************************/

#define MAXFIXSTRING 40

class FString {
friend ostream& operator << (ostream& os,FString& s);
protected:
  char Str[MAXFIXSTRING+1];
  size_t _Length;

  void CheckSize();

public:
  FString(const char far *src = "");
  FString(const char c);
  FString(const FString& src);
  FString(const DString& src);
  ~FString() { }
  const char far *c_str() const { return Str; }

  int Length() const { return _Length; }

  FString& operator =(const FString& src);
  FString& operator =(const char far *src);
  FString& operator =(const char c);
  BOOL operator ==(const char c) const {
    if (Length() != 1) return FALSE;
    return toupper(Str[0]) == toupper(c);
  }
  BOOL operator == (const FString& src) const {
    if (Length() != src.Length()) return FALSE;
    //return !stricmp(src.Str,Str);
    return fastcmp(src.Str,Str);
  }
  BOOL operator == (const char far *src) const {
    // return !stricmp(src,Str);
    return fastcmp(src,Str);
  }
  BOOL operator !=(const char c) const { return !(*this == c); }
  BOOL operator !=(const FString& src) const { return !(*this == src); }
  BOOL operator !=(const char far *src) const { return !(*this == src); }

  FString& operator +=(const FString& src) {
    _Length += src.Length();
    CheckSize();
    strcat(Str,src.Str);
    return *this;
  }

  FString operator +=(const char far *src) {
    _Length += strlen(src);
    CheckSize();
    strcat(Str,src);
    return *this;
  }

  FString operator +=(const char c) {
    _Length++;
    CheckSize();
    Str[_Length-1] = c;
    Str[_Length] = 0;
    return *this;
  }

  FString operator +(const FString& src);
  FString operator +(const char far *src);
  FString operator +(const char c);

  int NumLines();
  void ShrinkTo(int MaxLen);
  void Insert(char c,int pos=0);
  void Delete(int pos);
  char operator[] (int pos) {
    if ((pos < 0) || (pos > _Length))
      ErrStr << "Invalid string index" << ENDE;
    return Str[pos];
  }

  // removes everything to the right of the last period
  void RemoveFileExtension();

  void AllUpper() { strupr(Str); }
  void AllLower() { strlwr(Str); }
};

ostream& operator << (ostream& os,FString& s);

#endif

