/*
  TOKENIZE.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

  NOTE: This is the big bottleneck.  iostreams are *extremely* slow.  If I
	ever rewrite the tokenizer in assembly, the whole program will be
	*much* faster.  Believe it or not, CASM spends 80% of its time on
	the iostream::get() function.

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 

#include "tokenize.h"
#include "casm.h"

ostream& operator <<(ostream& os,TToken t) {
  return os << t.Text;
}

//--------------------------------------------------------------------------
TASMFileTokenizer::~TASMFileTokenizer() {
//--------------------------------------------------------------------------
  TFileData *ObjToDelete;
  while (FileStack) {
    ObjToDelete = FileStack;
    FileStack = FileStack->Next;
    delete ObjToDelete;
  }
}

//--------------------------------------------------------------------------
BOOL TASMFileTokenizer::_Eof() {
//--------------------------------------------------------------------------
  // first, find an open file
  if (!FileStack) return TRUE;
  while (FileStack->is.peek() == EOF) {
    // pop this record
    TotalLines += FileStack->LineNum;
    TFileData *ObjToDelete = FileStack;
    FileStack = FileStack->Next;
    delete ObjToDelete;
    if (!FileStack) return TRUE;
  }

  return FALSE;
}

//--------------------------------------------------------------------------
int TASMFileTokenizer::_GetWithComment() {
//--------------------------------------------------------------------------
startover:;
  if (_Eof()) return EOF;        // first, get to an open stream

  int ch = FileStack->GetAndCountLines();
  return ch;
}

//--------------------------------------------------------------------------
int TASMFileTokenizer::_Get() {  // gets, removing comments
//--------------------------------------------------------------------------
startover:;
  int ch = _GetWithComment();

  if (ch == ';') {
    ch = FileStack->is.peek();
    if (ch != '-') {
      // read to end of line:
      do {
	ch = FileStack->GetAndCountLines();
	if (FileStack->is.peek() == EOF) return EOF;
      } while (ch != '\n');
    } else { // block comment
      FileStack->GetAndCountLines();
      do {
	do { // read to next '-'
	  ch = FileStack->GetAndCountLines();
	  if (FileStack->is.peek() == EOF) {
	    FileStack->WriteFileLine(ErrStr);
	    ErrStr << "Unexpected end of file" << ENDE;
	  }

	  if ((ch == ';') && (FileStack->is.peek() == '-')) {
	    FileStack->WriteFileLine(ErrStr);
	    ErrStr << "\";-\" encountered inside a block comment" << ENDW;
	  }

	} while (ch != '-');
      } while (FileStack->is.peek() != ';');
      FileStack->GetAndCountLines(); // read the ';'
    }
    goto startover;
  }

  return ch;
}

//--------------------------------------------------------------------------
int TASMFileTokenizer::GetWithComment() {
//--------------------------------------------------------------------------
  if (TokenPeeked) {
    PeekToken.WriteFileLine(ErrStr);
    ErrStr << "PeekWithComment error" << PROGBUG;
  }
  if (CharPeeked) {
    CharPeeked = FALSE;
    return PeekChar;
  }
  return _GetWithComment();
}

//--------------------------------------------------------------------------
int TASMFileTokenizer::Get() {
//--------------------------------------------------------------------------
  if (TokenPeeked) {
    PeekToken.WriteFileLine(ErrStr);
    ErrStr << "PeekWithComment error" << PROGBUG;
  }
  if (CharPeeked) {
    CharPeeked = FALSE;
    if (FileStack) FileStack->PutBackAndCountLines(PeekChar);
  }
  return _Get();
}

//--------------------------------------------------------------------------
int TASMFileTokenizer::PeekWithComment() {
//--------------------------------------------------------------------------
  if (TokenPeeked) {
    PeekToken.WriteFileLine(ErrStr);
    ErrStr << "PeekWithComment error" << PROGBUG;
  }
  if (CharPeeked) return PeekChar;

  // if nothing peeked already, get something:
  PeekChar = _GetWithComment();
  if (PeekChar != EOF) CharPeeked = TRUE;
  return PeekChar;
}

int TASMFileTokenizer::Peek() {
  if (TokenPeeked) {
    PeekToken.WriteFileLine(ErrStr);
    ErrStr << "PeekWithComment error" << PROGBUG;
  }

startover:;

  if (CharPeeked) {
    if (PeekChar == ';') {  // un-peek the ';'
      if (FileStack) FileStack->PutBackAndCountLines(PeekChar);
      CharPeeked = FALSE;
      goto startover;
    }
    return PeekChar;
  }

  // if nothing peeked already, get something:
  PeekChar = _Get();
  if (PeekChar != EOF) CharPeeked = TRUE;
  return PeekChar;
}


//--------------------------------------------------------------------------
BOOL TASMFileTokenizer::IsSymbol(int c) {
//--------------------------------------------------------------------------
  if ((c >= '!') && (c <= '/')) return TRUE;
  if ((c >= ':') && (c <= '@')) return TRUE;
  if ((c >= '[') && (c <= '^')) return TRUE;
  if ((c >= '{') && (c <= '~')) return TRUE;
  return FALSE;
}

//--------------------------------------------------------------------------
void TASMFileTokenizer::SkipWhite() {
//--------------------------------------------------------------------------
  while (IsWhite(Peek())) Get();
}

//--------------------------------------------------------------------------
TToken TASMFileTokenizer::_GetNextToken() {
//--------------------------------------------------------------------------
  FString text;

  int c;

  SkipWhite();

  long TokenLine;
  if (FileStack) TokenLine = FileStack->LineNum;
  else TokenLine = 0;

  c = Get();

  if (c == EOF) return TToken();

  if (IsSymbol(c)) {
    if (FileStack) return TToken(FString((char)c),FileStack->Filename,
      TokenLine);
    else return TToken(FString((char)c));
  }

  if (isgraph(c)) {
    text = c;

    int len = 0;
    do {
      c = Peek();
      if (c == EOF) break;
      if (IsWhite(c)) break;
      if (IsSymbol(c)) break;
      Get(); // actually read c
      if (!isgraph(c)) {
	if (FileStack) FileStack->WriteFileLine(ErrStr);
	ErrStr << "Invalid character encountered in identifier: ";
	if isprint(c) ErrStr << '\'' << c << '\'';
	else ErrStr << "'\\" << (unsigned) c << '\'';
	ErrStr << ENDE;
      }
      text += c;
      if (++len > 32) {
	if (FileStack) FileStack->WriteFileLine(ErrStr);
	ErrStr << "Identifier exceeds maximum length" << ENDE;
      }
    } while(1);

    if (FileStack) return TToken(text,FileStack->Filename,TokenLine);
    else return TToken(text,"",0);
  }

  if (FileStack) FileStack->WriteFileLine(ErrStr);
  ErrStr << "Invalid character encountered in file: ";
  ErrStr.QuoteCharacter(c);
  ErrStr << ENDE;

  return TToken();  // unreachable code
}

//--------------------------------------------------------------------------
TToken TASMFileTokenizer::GetNextToken() {
//--------------------------------------------------------------------------
  if (TokenPeeked) {
    TokenPeeked = FALSE;
    return PeekToken;
  }
  return _GetNextToken();
}

//--------------------------------------------------------------------------
TToken TASMFileTokenizer::PeekNextToken() {
//--------------------------------------------------------------------------
  if (TokenPeeked) {
    return PeekToken;
  }
  PeekToken = _GetNextToken();
  TokenPeeked = TRUE;
  return PeekToken;
}

//--------------------------------------------------------------------------
BOOL TASMFileTokenizer::SkipTo(const FString& token) {
//--------------------------------------------------------------------------
  const char *searchstr = token.c_str();
  const char *curp = searchstr;
  char cur;

  if (!*curp) return TRUE;

  do {
    if (Eof()) return FALSE;

    cur = Get();

    if (IsWhite(cur) || (toupper(cur) != toupper(*curp))) {
      curp = searchstr;
      continue;
    }

    ++curp;
    if (!*curp) return TRUE;
  } while (1);
}

//--------------------------------------------------------------------------
DString TASMFileTokenizer::ReadToDelim(const char *delim) {
//--------------------------------------------------------------------------
  if (Eof()) return DString();

  char *check;
  DString buf;
  do {
    int c = PeekWithComment();

    check = (char *) delim;  // scan delim for c
    if (c == EOF) return buf;

    while (*check) {
      if (*check == c) return buf;  // hit delim
      check++;
    }


    GetWithComment(); // actually read c
    buf += (char) c;
  } while(1);
}

