/* 
  TOKENIZE.H - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#ifndef _TOKENIZE_H
#define _TOKENIZE_H

#include <fstream.h>

#include "general.h"
#include "petestr.h"

struct TToken {
  FString Text;
  FString Filename;
  unsigned long LineNum;

  TToken(const FString& _Text,const FString& _Filename = "",
    unsigned long _LineNum = 0)
    : Text(_Text), Filename(_Filename) { LineNum = _LineNum; }
  TToken() { LineNum = 0; }

  TToken& operator =(TToken &newval) {
    Text = newval.Text;
    Filename = newval.Filename;
    LineNum = newval.LineNum;
    return *this;
  }

  BOOL operator ==(const char *str) { return Text == str; }
  BOOL operator ==(const char c) { return Text == c; }
  BOOL operator ==(const FString& s) { return Text == s; }
  BOOL operator ==(const TToken& t) { return Text == t.Text; }
  BOOL operator !=(const char *str) { return Text != str; }
  BOOL operator !=(const char c) { return Text != c; }
  BOOL operator !=(const FString& s) { return Text != s; }
  BOOL operator !=(const TToken& t) { return Text != t.Text; }

  BOOL isnull() { return !Text.Length(); }

  TErrStr& WriteFileLine(TErrStr& es) {
    return es << '[' << Filename << ':' << LineNum << "] ";
  }
};

ostream& operator <<(ostream& os,TToken t);

class TASMFileTokenizer {
protected:
  struct TFileData {
    ifstream is;
    DString Filename;
    long LineNum;

    int GetAndCountLines() {
      int ch = is.get();
      if (ch == '\n')
	LineNum++;
      return ch;
    }
    void PutBackAndCountLines(int ch) {
      if (ch == '\n') LineNum--;
      is.putback(ch);
    }
    BOOL operator ==(TFileData &d) {
      return &d == this;
    }

    TErrStr& WriteFileLine(TErrStr& es) {
      return es << '[' << Filename << ':' << LineNum << "] ";
    }

    TFileData *Next;
  };

  TFileData *FileStack;

  // these do not support peeking:
  BOOL _Eof();
  int  _GetWithComment();  // pulls a character from file, processing
		// "@" directives, and counting lines
  int  _Get();  // gets, removing comments

  TToken PeekToken;    // used by PeekNextToken()
  BOOL TokenPeeked;

  int PeekChar;
  BOOL CharPeeked;

  unsigned long TotalLines;  // total lines tokenized so far

public:
  TASMFileTokenizer(const DString& startfile) {
    FileStack = 0;
    TotalLines = 0;
    TokenPeeked = FALSE;
    CharPeeked = FALSE;
    Open(startfile);
  }
  ~TASMFileTokenizer();

  void Open(const DString& Filename) {  // push a new file on the stack
    TFileData *d = new TFileData;
    d->LineNum = 1;
    d->Filename = Filename;
    d->Filename.AllUpper();
    d->is.open(d->Filename.c_str());
    if (d->is.fail())
      ErrStr << "Unable to open \"" << d->Filename.c_str() << '\"' << ENDE;
    d->is.clear();    // prevent false alarms

    // push this on the stack:
    d->Next = FileStack;
    FileStack = d;
  }

  BOOL Eof() { if (CharPeeked | TokenPeeked) return FALSE;
	       else return _Eof(); }
  int GetWithComment();
  int Get();
  int PeekWithComment();
  int Peek();

  BOOL IsWhite(int c) { return isspace(c); }

  BOOL IsSymbol(int c);

  void SkipWhite();

  TToken GetNextToken();
  TToken PeekNextToken();

  BOOL SkipTo(const FString& token);  // returns FALSE if EOF encountered

  // Reads up to until it reaches one of the characters in delim.  Does not extract that character
  DString ReadToDelim(const char *delim);

  TErrStr& WriteFileLine(TErrStr& es) {
    return es << '[' << FileStack->Filename << ':' << FileStack->LineNum << "] ";
  }

  // returns total lines tokenized so far, not counting lines in currently
  // open input stream.  (See Eof())
  unsigned long GetTotalLines() { return TotalLines; }

protected:
  TToken _GetNextToken();  // doesn't support peeking
};

#endif
